import logging
from pathlib import Path
from typing import Sequence

import pandas as pd

from duqtools.api import ImasHandle

from ..config import cfg
from ..merge import _merge, _resolve_variables
from ..models import Job, Locations
from ..operations import add_to_op_queue, op_queue

logger = logging.getLogger(__name__)


@add_to_op_queue('Writing csv', '{fname}')
def _write_data_csv(handles: dict[str, ImasHandle], fname: str = 'data.csv'):
    """Write handles to a data.csv file."""
    df = pd.DataFrame.from_dict(handles, orient='index')
    df.to_csv(fname)


def merge(force: bool, var_names: Sequence[str], **kwargs):
    cwd = Path.cwd()

    variables = _resolve_variables(var_names)

    config_files = cwd.glob('**/duqtools.yaml')

    target_handles = {}

    for config_file in config_files:
        run_name = config_file.parent.name

        cfg.parse_file(config_file)

        config_dir = config_file.parent

        runs = Locations(config_dir).runs
        runs = (run for run in runs if Job(run.dirname).is_completed)
        handles = (ImasHandle.parse_obj(run.data_out) for run in runs)
        handles = [handle for handle in handles if handle.exists()]

        if not handles:
            op_queue.warning(run_name, 'No data to merge.')
            continue

        op_queue.info(run_name,
                      extra_description=f'Merging {len(handles)} datasets')

        template_data = handles[0]

        target_data = template_data.copy()
        target_data.user = str(cfg.create.runs_dir / 'imasdb')
        target_handles[f'{run_name}_merged'] = target_data.dict()

        _merge(
            variables=variables,
            handles=handles,
            target=target_data,
            template=template_data,
            force=force,
        )

    _write_data_csv(target_handles, fname='merge_data.csv')



from ..setup import substitute_templates
from ..utils import read_imas_handles_from_file


def setup(*, template_file, input_file, force, **kwargs):
    """Setup large scale validation runs for template."""
    if not input_file:
        raise OSError('Input file not defined.')

    handles = read_imas_handles_from_file(input_file)

    substitute_templates(handles=handles,
                         template_file=template_file,
                         force=force)



import logging
import os
import shutil
import stat
import subprocess as sp
import sys
from collections.abc import Sequence
from pathlib import Path
from typing import Any, List, Optional

from jetto_tools import config, jset, lookup, namelist, template
from jetto_tools import job as jetto_job
from jetto_tools.template import _EXTRA_FILE_REGEXES

from ..config import Config, cfg
from ..ids import ImasHandle
from ..models import AbstractSystem, Job, Locations
from ..operations import add_to_op_queue
from ..schema import JettoVar
from ._batchfile import write_array_batchfile as _write_array_batchfile
from ._batchfile import write_batchfile as _write_batchfile
from ._jettovar_to_json import jettovar_to_json

if sys.version_info < (3, 10):
    from importlib_resources import files
else:
    from importlib.resources import files

logger = logging.getLogger(__name__)

lookup_file = files('duqtools.data') / 'jetto_tools_lookup.json'
jetto_lookup = lookup.from_file(lookup_file)


class BaseJettoSystem(AbstractSystem):
    """System that can be used to create runs for jetto.

    This system can submit to various backends like docker, prominence
    and the gateway.
    """

    @staticmethod
    def get_runs_dir() -> Path:
        path = Locations().jruns_path
        runs_dir = cfg.create.runs_dir  # type: ignore
        if not runs_dir:
            abs_cwd = str(Path.cwd().resolve())
            abs_jruns = str(path.resolve())
            # Check if jruns is parent dir of current dir
            if abs_cwd.startswith(abs_jruns):
                runs_dir = Path()
            else:  # jruns is somewhere else
                count = 0
                while True:  # find the next free folder
                    runs_dir = f'duqtools_experiment_{count:04d}'
                    if not (path / runs_dir).exists():
                        break
                    count = count + 1
        return path / runs_dir

    @staticmethod
    @add_to_op_queue('Writing new batchfile', '{run_dir.name}', quiet=True)
    def write_batchfile(run_dir: Path, cfg: Config):
        jetto_jset = jset.read(run_dir / 'jetto.jset')
        _write_batchfile(run_dir, jetto_jset, tag=cfg.tag)

    @staticmethod
    def submit_job(job: Job):
        # Make sure we get a new correct status
        if (job.dir / 'jetto.status').exists():
            os.remove(job.dir / 'jetto.status')

        submit_system = job.cfg.submit.submit_system

        if submit_system == 'slurm':
            submit = JettoSystem.submit_slurm
        elif submit_system == 'docker':
            submit = JettoSystem.submit_docker
        elif submit_system == 'prominence':
            submit = JettoSystem.submit_prominence
        else:
            raise NotImplementedError(f'submission type {submit_system}'
                                      ' not implemented')

        submit(job)

    @staticmethod
    def submit_slurm(job: Job):
        if not job.has_submit_script:
            raise FileNotFoundError(job.submit_script)

        submit_cmd = job.cfg.submit.submit_command.split()
        cmd: list[Any] = [*submit_cmd, str(job.submit_script)]

        logger.info(f'submitting script via slurm {cmd}')

        ret = sp.run(cmd, check=True, capture_output=True)
        logger.info('submission returned: ' + str(ret.stdout))
        with open(job.lockfile, 'wb') as f:
            f.write(ret.stdout)

    @staticmethod
    def submit_docker(job: Job):
        jetto_template = template.from_directory(job.dir)
        jetto_config = config.RunConfig(jetto_template)
        jetto_manager = jetto_job.JobManager()
        extra_volumes = {
            job.dir.parent / 'imasdb': {
                'bind': '/opt/imas/shared/imasdb',
                'mode': 'rw'
            },
        }

        os.environ['RUNS_HOME'] = os.getcwd()
        os.environ['JINTRAC_IMAS_BACKEND'] = 'MDSPLUS'
        container = jetto_manager.submit_job_to_docker(
            jetto_config,
            job.dir,
            image=job.cfg.submit.docker_image,
            extra_volumes=extra_volumes)
        job.lockfile.touch()
        with open(job.lockfile, 'w') as f:
            f.write(container.name)

    @staticmethod
    def submit_prominence(job: Job):
        jetto_template = template.from_directory(job.dir)
        jetto_config = config.RunConfig(jetto_template)
        jetto_manager = jetto_job.JobManager()

        os.environ['RUNS_HOME'] = os.getcwd()
        _ = jetto_manager.submit_job_to_prominence(jetto_config, job.dir)

    @staticmethod
    def submit_array(jobs: Sequence[Job], max_jobs: int):
        if jobs[0].cfg.submit.submit_system == 'slurm':
            JettoSystem.submit_array_slurm(jobs, max_jobs)
        else:
            raise NotImplementedError(
                'array submission type {jobs[0].cfg.submit.submit_system}'
                ' not implemented')

    @staticmethod
    @add_to_op_queue('Submit single array job', 'duqtools_slurm_array.sh')
    def submit_array_slurm(jobs: Sequence[Job], max_jobs: int):
        for job in jobs:
            job.lockfile.touch()

        logger.info('writing duqtools_slurm_array.sh file')
        _write_array_batchfile(jobs, max_jobs)

        submit_cmd = jobs[0].cfg.submit.submit_command.split()
        cmd: list[Any] = [*submit_cmd, 'duqtools_slurm_array.sh']

        logger.info(f'Submitting script via: {cmd}')

        ret = sp.run(cmd, check=True, capture_output=True)
        logger.info('submission returned: ' + str(ret.stdout))

        for job in jobs:
            with open(job.lockfile, 'wb') as f:
                f.write(ret.stdout)

    @staticmethod
    def _apply_patches_to_template(jetto_template: template.Template):
        """Apply settings that are necessary for duqtools to function."""
        # Force output of IDS data
        # https://github.com/duqtools/duqtools/issues/343
        jetto_template.jset._settings['JobProcessingPanel.selIdsRunid'] = True

    @staticmethod
    @add_to_op_queue('Copying template to', '{target_drc}', quiet=True)
    def copy_from_template(source_drc: Path, target_drc: Path):
        jetto_jset = jset.read(source_drc / 'jetto.jset')
        jetto_namelist = namelist.read(source_drc / 'jetto.in')

        jetto_sanco = None
        if (source_drc / 'jetto.sin').exists():
            jetto_sanco = namelist.read(source_drc / 'jetto.sin')

        all_files = os.listdir(source_drc)

        jetto_extra: List[str] = []
        for regex in _EXTRA_FILE_REGEXES:
            jetto_extra.extend(filter(regex.match, all_files))

        if 'jetto.restart' in jetto_extra or 'jetto.srestart' in jetto_extra:
            raise OSError(
                'Template contains `jetto.restart` and/or `jetto.srestart` files. '
                'Please remove these before running duqtools again. '
                'More info: https://github.com/duqtools/duqtools/issues/498')

        jetto_extra = [str(source_drc / file) for file in jetto_extra]

        jetto_template = template.Template(jset=jetto_jset,
                                           namelist=jetto_namelist,
                                           lookup=jetto_lookup,
                                           sanco_namelist=jetto_sanco,
                                           extra_files=jetto_extra)

        JettoSystem._apply_patches_to_template(jetto_template)

        jetto_config = config.RunConfig(jetto_template)

        jetto_config.export(target_drc)
        lookup.to_file(jetto_lookup, target_drc /
                       'lookup.json')  # TODO, this should be copied as well

        for filename in (
                'rjettov',
                'utils_jetto',
        ):
            src = source_drc / filename
            dst = target_drc / filename
            shutil.copyfile(src, dst)
            dst.chmod(dst.stat().st_mode | stat.S_IXUSR)

    @staticmethod
    def imas_from_path(template_drc: Path) -> ImasHandle:
        jetto_jset = jset.read(template_drc / 'jetto.jset')

        return ImasHandle(
            db=jetto_jset['SetUpPanel.idsIMASDBMachine'],  # type: ignore
            user=jetto_jset['SetUpPanel.idsIMASDBUser'],  # type: ignore
            run=jetto_jset['SetUpPanel.idsIMASDBRunid'],  # type: ignore
            shot=jetto_jset['SetUpPanel.idsIMASDBShot'])  # type: ignore

    @staticmethod
    @add_to_op_queue('Updating imas locations of', '{run}', quiet=True)
    def update_imas_locations(run: Path, inp: ImasHandle, out: ImasHandle):
        jetto_template = template.from_directory(run)
        jetto_config = config.RunConfig(jetto_template)

        jetto_config['user_in'] = inp.user
        jetto_config['machine_in'] = inp.db
        jetto_config['shot_in'] = inp.shot
        jetto_config['run_in'] = inp.run

        jetto_config['machine_out'] = out.db
        jetto_config['shot_out'] = out.shot
        jetto_config['run_out'] = out.run

        jetto_config.export(run)  # Just overwrite the poor files

    @staticmethod
    def set_jetto_variable(run: Path,
                           key: str,
                           value,
                           variable: Optional[JettoVar] = None):
        jetto_template = template.from_directory(run)

        if variable:
            extra_lookup = lookup.from_json(jettovar_to_json(variable))
            jetto_template._lookup.update(extra_lookup)

        jetto_config = config.RunConfig(jetto_template)

        if key == 't_start':
            jetto_config.start_time = value
        elif key == 't_end':
            jetto_config.end_time = value
        else:
            jetto_config[key] = value

        jetto_config.export(run)  # Just overwrite the poor files


class JettoSystemV210921(BaseJettoSystem):
    """System that can be used to create runs for jetto.

    The backend that is  assumed is jetto-v210921.

    This system can submit to various backends like docker, prominence
    and the gateway.

    ```yaml title="duqtools.yaml"
    system: jetto-v210921
    submit:
      submit_system: prominence
    ```
    """

    @staticmethod
    def get_data_in_handle(
        *,
        dirname: Path,
        source: ImasHandle,
        seq_number: int,
        options,
    ):
        """Get handle for data input."""
        return ImasHandle(
            user=options.user,
            db=options.imasdb,
            shot=source.shot,
            run=options.run_in_start_at + seq_number,
        )

    @staticmethod
    def get_data_out_handle(
        *,
        dirname: Path,
        source: ImasHandle,
        seq_number: int,
        options,
    ):
        """Get handle for data output."""
        return ImasHandle(
            user=options.user,
            db=options.imasdb,
            shot=source.shot,
            run=options.run_out_start_at + seq_number,
        )


class JettoSystemV220922(BaseJettoSystem):
    """System that can be used to create runs for jetto.

    The backend that is  assumed is jetto-v220922. The most important
    difference with v210921 is that the IMAS data is handled locally
    instead of via a public `imasdb`

    This system can submit to various backends like docker, prominence
    and the gateway.

    ```yaml title="duqtools.yaml"
    system: jetto-v220922
    submit:
      submit_system: docker
    ```
    """

    @staticmethod
    def get_data_in_handle(
        *,
        dirname: Path,
        source: ImasHandle,
        seq_number: int,
        options,
    ):
        relative_location: Optional[str] = str(
            os.path.relpath((dirname / 'imasdb').resolve()))
        if relative_location:
            if relative_location.startswith('..'):
                relative_location = None
        """Get handle for data input."""
        return ImasHandle(user=str((dirname / 'imasdb').resolve()),
                          db=source.db,
                          shot=source.shot,
                          run=1,
                          relative_location=relative_location)

    @staticmethod
    def get_data_out_handle(
        *,
        dirname: Path,
        source: ImasHandle,
        seq_number: int,
        options,
    ):
        relative_location: Optional[str] = str(
            os.path.relpath((dirname / 'imasdb').resolve()))
        if relative_location:
            if relative_location.startswith('..'):
                relative_location = None
        """Get handle for data output."""
        return ImasHandle(user=str((dirname / 'imasdb').resolve()),
                          db=source.db,
                          shot=source.shot,
                          run=2,
                          relative_location=relative_location)


JettoSystem = JettoSystemV220922



from collections import deque
from pathlib import Path
from typing import Deque, Sequence

from ..config import cfg
from ..models import Job, Locations
from ..submit import (
    SubmitError,
    job_array_submitter,
    job_scheduler,
    job_submitter,
    lockfile_ok,
    status_file_ok,
)


def submit(*, array, force, max_jobs, schedule, status_filter: Sequence[str],
           **kwargs):
    """Submit nested duqtools configs.

    Parameters
    ----------
    force : bool
        Force the submission even in the presence of lockfiles
    max_jobs : int
        Maximum number of jobs to submit at once
    schedule : bool
        Schedule `max_jobs` to run at once, keeps the process alive until
        finished.
    status_filter : list[str]
        Only submit jobs with this status.
    """
    cwd = Path.cwd()

    config_files = cwd.glob('**/duqtools.yaml')

    jobs: list[Job] = []

    for config_file in config_files:
        cfg.parse_file(config_file)

        if not cfg.submit:
            raise SubmitError(
                f'Submit field required in config file: {config_file}')

        config_dir = config_file.parent

        jobs.extend(Job(run.dirname) for run in Locations(config_dir).runs)

    job_queue: Deque[Job] = deque()

    for job in jobs:
        status = job.status()

        if status_filter and (status not in status_filter):
            continue
        if not status_file_ok(job, force=force):
            continue
        if not lockfile_ok(job, force=force):
            continue
        job_queue.append(job)

    submitter = job_scheduler if schedule else job_submitter
    submitter = job_array_submitter if array else submitter
    submitter(job_queue, max_jobs=max_jobs)



from __future__ import annotations

import logging
from typing import Sequence

from .config import var_lookup
from .ids import ImasHandle, merge_data
from .operations import op_queue
from .schema import IDSVariableModel
from .utils import read_imas_handles_from_file

logger = logging.getLogger(__name__)
info, debug = logger.info, logger.debug


def _resolve_variables(var_names: Sequence[str]) -> list[IDSVariableModel]:
    """Looks up variables if specified, if empty return all variables."""
    idsvar_lookup = var_lookup.filter_type('IDS-variable')

    if not var_names:
        variables = list(idsvar_lookup.values())
        op_queue.info(description='Merging all known variables')
    else:
        variables = list(idsvar_lookup[name] for name in var_names)
        for variable in variables:
            op_queue.info(description='Variable for merge',
                          extra_description=f'{variable.name}')

    return variables


def _merge(*,
           handles: Sequence[ImasHandle],
           template: ImasHandle,
           target: ImasHandle,
           variables: Sequence[IDSVariableModel],
           force: bool = False):
    """Merge mas data.

    Parameters
    ----------
    handles : Sequence[ImasHandle]
        These are the imas handles that will be merged.
    template : ImasHandle
        This handle is the template that is used for the new data.
    target : ImasHandle
        This handle is the target location where the merged data will be stored.
    variables : Sequence[Variable
        These are the IDS variables to be merged.
    force : bool
        Force overwriting existing files.
    """
    for handle in handles:
        logger.debug('Source for merge %s', handle)

    op_queue.info(description='Template for merge',
                  extra_description=f'{template}')

    if target.exists() and not force:
        op_queue.add_no_op(description='Abort merge',
                           extra_description=f'{target} already exists, '
                           'use --force to overwrite')
        return

    template.copy_data_to(target)

    merge_data(handles, target, variables)


def merge(*, target: str, template: str, handles: list[str],
          input_files: list[str], var_names: list[str], force: bool, **kwargs):
    """Merge as many data as possible."""
    template = ImasHandle.from_string(template)
    target = ImasHandle.from_string(target)

    handles = [ImasHandle.from_string(handle) for handle in handles]
    for file in input_files:
        handles = handles + list(read_imas_handles_from_file(file).values())

    handles = list(set(handles))  # Remove duplicate handles

    variables = _resolve_variables(var_names)

    _merge(
        handles=handles,
        template=template,
        target=target,
        variables=variables,
    )



"""Function to create llcmd file."""
from __future__ import annotations

import stat
from os.path import commonpath
from pathlib import Path
from typing import TYPE_CHECKING, Sequence

import jetto_tools

from ..models import Locations

if TYPE_CHECKING:
    from ..models import Job


def write_batchfile(run_dir: Path,
                    jset: jetto_tools.jset.JSET,
                    tag: str | None = None):
    """Write batchfile (`.llcmd`) to start jetto.

    Parameters
    ----------
    run_dir : Path
        Directory to place batch file into.
    jset: JSET
        Jetto settings object (from jetto-pythontools)
    tag : str | None
        Tag the job, defaults to 'jetto'.
    """
    if not tag:
        tag = 'jetto'

    llcmd_path = run_dir / '.llcmd'

    rjettov_path = run_dir / 'rjettov'
    rel_path = run_dir.resolve().relative_to(Locations().jruns_path.resolve())

    build_name = jset['JobProcessingPanel.name']
    build_user_name = jset['JobProcessingPanel.userid']
    machine_number = jset['JobProcessingPanel.machineNumber']
    num_proc = jset['JobProcessingPanel.numProcessors']
    wall_time = jset['JobProcessingPanel.wallTime']

    string = f"""#!/bin/sh
#SBATCH -J duqtools.{tag}.{run_dir.name}
#SBATCH -i /dev/null
#SBATCH -o {run_dir / 'll.out'}
#SBATCH -e {run_dir / 'll.err'}
#SBATCH -p {machine_number}

#SBATCH -N 1
#SBATCH -n {num_proc}
#SBATCH -t {wall_time}:00:00

cd {run_dir}
{rjettov_path} -S -I -p -xmpi -x64 {rel_path} \
{build_name} {build_user_name}"""

    with open(llcmd_path, 'w') as f:
        f.write(string)

    llcmd_path.chmod(llcmd_path.stat().st_mode | stat.S_IXUSR)


def write_array_batchfile(jobs: Sequence[Job], max_jobs: int):
    """Write array batchfile to start jetto runs.

    Parameters
    ----------
    jobs : Sequence[Job]
        List of jobs to run.
    max_jobs : int
        Maximum number of jobs to run at the same time.
    """
    common_dir = Path(commonpath(job.dir for job in jobs))  # type: ignore
    logs_dir = common_dir / 'logs'
    logs_dir.mkdir(exist_ok=True)

    # Get the first jobs submission script as a template
    lines = open(jobs[0].submit_script)
    sbatch_lines = (line for line in lines if line.startswith('#SBATCH'))
    option_lines = (line for line in sbatch_lines
                    if line.split()[1] not in ('-o', '-e', '-J'))
    options = ''.join(option_lines)

    # Append our own options, later options have precedence
    out_file = logs_dir / 'duqtools-%A_%a.out'
    err_file = logs_dir / 'duqtools-%A_%a.err'

    scripts = '\n'.join(f'    {job.submit_script}' for job in jobs)

    string = f"""#!/bin/sh
{options}
#SBATCH -o {out_file}
#SBATCH -e {err_file}
#SBATCH --array=0-{len(jobs)-1}%{max_jobs}
#SBATCH -J duqtools-array

scripts=(
{scripts}
)
echo executing ${{scripts[$SLURM_ARRAY_TASK_ID]}}
${{scripts[$SLURM_ARRAY_TASK_ID]}} || true
"""

    with open('duqtools_slurm_array.sh', 'w') as f:
        f.write(string)



# https://setuptools.pypa.io/en/latest/pkg_resources.html#workingset-objects
def fix_dependencies():
    import __main__
    __main__.__requires__ = [
        'jetto_tools>=1.8.8',
        'pydantic!=1.10.3',
        'pydantic!=1.10.4',
        'pydantic!=1.10.5',
        'scipy>=1.09',
        'jinja2>=3.0.0',
    ]
    import pkg_resources  # noqa


fix_dependencies()

__author__ = 'Stef Smeets'
__email__ = 's.smeets@esciencecenter.nl'
__version__ = '1.6.3'

import logging  # noqa
import warnings  # noqa

logging.basicConfig(level=logging.INFO)

# https://github.com/duqtools/duqtools/issues/264
warnings.filterwarnings(
    'ignore',
    'Explicit custom root behavior not yet implemented for pydantic_yaml')
